/*
	JSON parser & writer by Stuart Welch <afrowuk@afrowsoft.co.uk>
	v1.1.0.6 - 14th October 2016
*/

#include <windows.h>
#include "JSON.h"
#include "pluginapi.h"

#define IsWhitespace(c) c == TEXT(' ') || c == TEXT('\t') || c == TEXT('\r') || c == TEXT('\n')
#define ntohs(n) (((((unsigned short)(n) & 0xFF)) << 8) | (((unsigned short)(n) & 0xFF00) >> 8))
static TCHAR hex[] = TEXT("0123456789abcdef");

struct JSON_NODE* JSON_Create()
{
	return (struct JSON_NODE*)GlobalAlloc(GPTR, sizeof(struct JSON_NODE));
}

PTCHAR JSON_GetQuotedValue(struct JSON_NODE* pNode, const PTCHAR pszDefaultValue)
{
	return pNode && pNode->eType == JNT_QUOTED_VALUE ? pNode->pszValue : pszDefaultValue;
}

BOOL JSON_IsTrue(struct JSON_NODE* pNode)
{
	if (!pNode || !pNode->pszValue || lstrcmp(pNode->pszValue, TEXT("0")) == 0 || lstrcmpi(pNode->pszValue, TEXT("false")) == 0)
		return FALSE;
	return TRUE;
}

static void EatWhitespace(PTCHAR pszBuffer, int* piPos)
{
	while (IsWhitespace(pszBuffer[*piPos]))
		(*piPos)++;
}

static PTCHAR MakeCopy(PTCHAR pszCopyFrom, int iStart, int iEnd)
{
	PTCHAR pszCopy = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (iEnd - iStart + 2));
	if (pszCopy)
	{
		int i = 0;
		for (i = 0; iStart <= iEnd; i++, iStart++)
			pszCopy[i] = pszCopyFrom[iStart];
	}

	return pszCopy;
}

static BOOL IsChar(PTCHAR pszBuffer, int* piPos, TCHAR chExpected)
{
	EatWhitespace(pszBuffer, piPos);

	if (pszBuffer[*piPos] != chExpected)
		return FALSE;

	return TRUE;
}

/*static BOOL IsNumeric(PTCHAR pszBuffer)
{
	int cch = lstrlen(pszBuffer), i;
	if (cch == 0)
		return FALSE;

	if (pszBuffer[0] == TEXT('0'))
	{
		if (cch == 1)
			return TRUE;
		return FALSE;
	}

	for (i = 0; i < cch; i++)
		if (pszBuffer[i] < TEXT('0') || pszBuffer[i] > TEXT('9'))
			return FALSE;

	return TRUE;
}*/

static BOOL EatChar(PTCHAR pszBuffer, int* piPos, TCHAR chExpected)
{
	if (!IsChar(pszBuffer, piPos, chExpected))
		return FALSE;

	(*piPos)++;
	return TRUE;
}

static BOOL IsEscaped(PTCHAR pszBuffer, int iPos)
{
	return iPos > 0 && pszBuffer[iPos - 1] == TEXT('\\') && !IsEscaped(pszBuffer, iPos - 1);
}

static enum JSON_WORD_TYPE EatWord(PTCHAR pszBuffer, int* piPos, PTCHAR* ppszWord)
{
	int iStart, iEnd;

	EatWhitespace(pszBuffer, piPos);

	iStart = *piPos;
	iEnd = 0;
	*ppszWord = NULL;

	if (pszBuffer[*piPos] == TEXT('"'))
	{
		iStart++;

		while (TRUE)
		{
			(*piPos)++;
			if (pszBuffer[*piPos] == TEXT('"') && !IsEscaped(pszBuffer, *piPos) || pszBuffer[*piPos] == 0)
			{
				iEnd = *piPos - 1;
				if (pszBuffer[*piPos] == TEXT('"'))
					(*piPos)++;
				break;
			}
		}

		*ppszWord = MakeCopy(pszBuffer, iStart, iEnd);
		return JWT_STRING;
	}

	while (TRUE)
	{
		if (pszBuffer[*piPos] == TEXT(':') || pszBuffer[*piPos] == TEXT(',') || pszBuffer[*piPos] == TEXT('}') || pszBuffer[*piPos] == TEXT(']') || IsWhitespace(pszBuffer[*piPos]) || pszBuffer[*piPos] == 0)
		{
			iEnd = *piPos - 1;
			break;
		}
		(*piPos)++;
	}

	if (iStart <= iEnd)
	{
		*ppszWord = MakeCopy(pszBuffer, iStart, iEnd);
		return JWT_OTHER;
	}

	return JWT_NONE;
}

static struct JSON_NODE* EatNode(PTCHAR pszBuffer, int* piPos, BOOL bIsValue);

static struct JSON_NODE* EatNodeArray(PTCHAR pszBuffer, int* piPos)
{
	struct JSON_NODE* pNode = EatNode(pszBuffer, piPos, TRUE);
	if (pNode && EatChar(pszBuffer, piPos, TEXT(',')))
		pNode->pNext = EatNodeArray(pszBuffer, piPos);
	return pNode;
}

static struct JSON_NODE* EatNode(PTCHAR pszBuffer, int* piPos, BOOL bIsValue)
{
	struct JSON_NODE* pNode = JSON_Create();

	if (pNode)
	{
		if (EatChar(pszBuffer, piPos, TEXT('{')))
		{
			pNode->eType = JNT_NODE;
			pNode->pValue = EatNode(pszBuffer, piPos, FALSE);
			EatChar(pszBuffer, piPos, TEXT('}'));
		}
		else if (EatChar(pszBuffer, piPos, TEXT('[')))
		{
			pNode->eType = JNT_ARRAY;
			pNode->pValue = EatNodeArray(pszBuffer, piPos);
			EatChar(pszBuffer, piPos, TEXT(']'));
		}
		else
		{
			PTCHAR pszValue = NULL;
			enum JSON_WORD_TYPE eType = EatWord(pszBuffer, piPos, &pszValue);

			if (pszValue)
			{
				switch (eType)
				{
				case JWT_STRING:
					pNode->eType = JNT_QUOTED_VALUE;
					break;
				case JWT_OTHER:
					pNode->eType = JNT_VALUE;
					break;
				}
    
				if (eType != JWT_NONE)
				{
					// Node is just a value.
					if (bIsValue)
					{
						pNode->pszValue = pszValue;
					}
					// Node is a key: value pair.
					else if (EatChar(pszBuffer, piPos, TEXT(':')))
					{
						struct JSON_NODE* pChildNode = EatNode(pszBuffer, piPos, TRUE);
						if (pChildNode)
						{
							pNode->eType = pChildNode->eType;
							pNode->pNext = pChildNode->pNext;
							pNode->pValue = pChildNode->pValue;
							GlobalFree(pChildNode);
						}

						pNode->pszKey = pszValue;
					}
					// No key was given; use an empty string.
					else
					{
						pNode->pszKey = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR));
						pNode->pszValue = pszValue;
					}
				}
			}
			else
			{
				GlobalFree(pNode);
				return NULL;
			}
		}
    
		// Commas are allowed; eat next node.
		if (!bIsValue && EatChar(pszBuffer, piPos, TEXT(',')))
			pNode->pNext = EatNode(pszBuffer, piPos, FALSE);
	}

	return pNode;
}

static struct JSON_NODE* EatRoot(PTCHAR pszBuffer)
{
	int iPos = 0;
	return EatNode(pszBuffer, &iPos, FALSE);
}

static PTCHAR EscapeQuotes(PTCHAR pszStr)
{
	int nQuotes = 0, i, cchLen = lstrlen(pszStr);
	PTCHAR pszStrNew;
  
	for (i = 0; i < cchLen; i++)
		if (pszStr[i] == TEXT('"'))
			nQuotes++;

	pszStrNew = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (cchLen + nQuotes + 1));
	if (pszStrNew)
	{
		int j;
    
		for (i = 0, j = 0; i < cchLen; i++, j++)
		{
			if (pszStr[i] == TEXT('"') && !IsEscaped(pszStr, i))
			{
				pszStrNew[j] = TEXT('\\');
				j++;
			}

			pszStrNew[j] = pszStr[i];
		}
	}
  
	return pszStrNew;
}

void JSON_Delete(struct JSON_NODE* pNode, struct JSON_NODE* pPrev)
{
	if (pNode)
	{
		struct JSON_NODE* pNext;

		if (pPrev)
		{
			if (pPrev->pNext == pNode)
				pPrev->pNext = pNode->pNext;

			if ((pPrev->eType == JNT_NODE || pPrev->eType == JNT_ARRAY) && pPrev->pValue == pNode)
				pPrev->pValue = pNode->pNext;
		}

		switch (pNode->eType)
		{
		case JNT_NODE:
		case JNT_ARRAY:

			pNext = pNode->pValue;
			while (pNext)
			{
				struct JSON_NODE* pTemp = pNext->pNext;
				JSON_Delete(pNext, NULL);
				pNext = pTemp;
			}

			break;
		case JNT_VALUE:
		case JNT_QUOTED_VALUE:

			GlobalFree(pNode->pszValue);

			break;
		}
    
		if (pNode->pszKey)
			GlobalFree(pNode->pszKey);
		GlobalFree(pNode);
	}
}

static PCHAR HandleUTF8Input(PBYTE pbBuffer, int* pcbBuffer)
{
	if (pbBuffer[0] == 0xEF && pbBuffer[1] == 0xBB && pbBuffer[2] == 0xBF) // UTF-8 sig
	{
		if (pcbBuffer)
			*pcbBuffer -= 3;
		return (PCHAR)(pbBuffer + 3);
	}

	return (PCHAR)pbBuffer;
}

static PWCHAR HandleUTF16Input(PBYTE pbBuffer, int* pcchBuffer)
{
	if (pbBuffer[0] == 0xFE && pbBuffer[1] == 0xFF) // UTF-16BE BOM
	{
		int i;
		PWCHAR pwszBuffer = (PWCHAR)(pbBuffer + 2);
		for (i = 0; pwszBuffer[i] != L'\0'; i++)
			pwszBuffer[i] = ntohs(pwszBuffer[i]);
		if (pcchBuffer)
			*pcchBuffer -= 2;
		return pwszBuffer;
	}

	if (pbBuffer[0] == 0xFF && pbBuffer[1] == 0xFE && (pbBuffer[2] != 0x00 || pbBuffer[3] != 0x00)) // UTF-16LE BOM
	{
		if (pcchBuffer)
			*pcchBuffer -= 2;
		return (PWCHAR)(pbBuffer + 2);
	}

	return (PWCHAR)pbBuffer;
}

static struct JSON_NODE* Parse(PBYTE pbSource, enum JSON_SET_FLAGS eFlags)
{
	struct JSON_NODE* pNode = NULL;
	PBYTE pbBuffer = NULL;
	DWORD dwSize = 0;

	if (eFlags & JSF_IS_FILE)
	{
		HANDLE hFile = CreateFile((PTCHAR)pbSource, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile != INVALID_HANDLE_VALUE)
		{
			dwSize = GetFileSize(hFile, NULL);
			if (dwSize > 0)
			{
				pbBuffer = (PBYTE)GlobalAlloc(GPTR, dwSize + ((eFlags & JSF_IS_UNICODE) ? sizeof(WCHAR) : sizeof(CHAR)));
				if (pbBuffer)
					ReadFile(hFile, pbBuffer, dwSize, &dwSize, NULL);
			}
			CloseHandle(hFile);
		}
	}
	else
	{
		pbBuffer = pbSource;
		dwSize = (eFlags & JSF_IS_UNICODE) ? lstrlenW((PWCHAR)pbSource) : lstrlenA((PCHAR)pbSource);
	}

	if (pbBuffer)
	{
#ifdef UNICODE
		if (eFlags & JSF_IS_UNICODE)
		{
			pNode = EatRoot(HandleUTF16Input(pbBuffer, NULL));
		}
		else
		{
			int cbBuffer = dwSize;
			PWCHAR pwszBuffer = JSON_ToUnicode(HandleUTF8Input(pbBuffer, &cbBuffer), &cbBuffer);
			if (pwszBuffer)
			{
				pNode = EatRoot(pwszBuffer);
				GlobalFree(pwszBuffer);
			}
		}
#else
		if (eFlags & JSF_IS_UNICODE)
		{
			int cchBuffer = dwSize / sizeof(WCHAR);
			PCHAR pszBuffer = JSON_FromUnicode(HandleUTF16Input(pbBuffer, &cchBuffer), &cchBuffer);
			if (pszBuffer)
			{
				pNode = EatRoot(pszBuffer);
				GlobalFree(pszBuffer);
			}
		}
		else
		{
			pNode = EatRoot(HandleUTF8Input(pbBuffer, NULL));
		}
#endif

		if (eFlags & JSF_IS_FILE)
			GlobalFree(pbBuffer);
	}

	return pNode;
}

int JSON_Count(struct JSON_NODE* pNode)
{
	int i = 0;

	if (pNode)
	{
		pNode = pNode->pValue;
		for (; pNode != NULL; i++)
			pNode = pNode->pNext;
	}

	return i;
}

struct JSON_NODE* JSON_Get(struct JSON_NODE* pNode, PTCHAR pszKey, BOOL bKeyIsIndex)
{
	return JSON_Next(&pNode, pszKey, bKeyIsIndex, FALSE);
}

struct JSON_NODE* JSON_GetEx(struct JSON_NODE* pNode, PTCHAR pszKey, BOOL bKeyIsIndex, BOOL bCreate)
{
	return JSON_Next(&pNode, pszKey, bKeyIsIndex, bCreate);
}

struct JSON_NODE* JSON_Next(struct JSON_NODE** ppNode, PTCHAR pszKey, BOOL bKeyIsIndex, BOOL bCreate)
{
	if (ppNode)
	{
		struct JSON_NODE* pParent = *ppNode;
		struct JSON_NODE* pNext = NULL;

		pszKey = EscapeQuotes(pszKey);
		if (pszKey)
		{
			// We can only get a child node if the parent is also a node or array.
			if (pParent->eType == JNT_NODE || pParent->eType == JNT_ARRAY)
			{
				pNext = pParent->pValue;
      
				// Get the child node by index.
				if (bKeyIsIndex)
				{
					int i, j = myatoi(pszKey);

					// Negative index?
					if (j < 0)
						j = JSON_Count(pParent) + j;

					for (i = 0; i < j && pNext != NULL; i++)
					{
						*ppNode = pNext;
						pNext = pNext->pNext;
					}
				}
				// Get the child node by key.
				else if (pParent->eType == JNT_NODE)
				{
					while (pNext != NULL && !(!pNext->pszKey && !*pszKey || lstrcmp(pNext->pszKey, pszKey) == 0))
					{
						*ppNode = pNext;
						pNext = pNext->pNext;
					}
				}
				// Seek to the end of the array if we're going to add a new node element.
				else if (bCreate)
				{
					while (pNext != NULL)
					{
						*ppNode = pNext;
						pNext = pNext->pNext;
					}
				}
				// Otherwise find an array element that matches.
				else
				{
					while (pNext != NULL)
					{
						if ((pNext->eType == JNT_VALUE || pNext->eType == JNT_QUOTED_VALUE) && lstrcmp(pNext->pszValue, pszKey) == 0)
							break;

						*ppNode = pNext;
						pNext = pNext->pNext;
					}
				}

				// No existing child node found and the caller wants to create one.
				if (!pNext && bCreate)
				{
					struct JSON_NODE* pNew = (struct JSON_NODE*)GlobalAlloc(GPTR, sizeof(struct JSON_NODE));
					pNew->eType = JNT_NODE;

					// Adding a new node to an array.
					if (pParent->eType == JNT_ARRAY && !bKeyIsIndex)
					{
						pNew->pValue = (struct JSON_NODE*)GlobalAlloc(GPTR, sizeof(struct JSON_NODE));
						pNew->pValue->eType = JNT_NODE;

						if (*pszKey)
						{
							pNew->pValue->pszKey = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (lstrlen(pszKey) + 1));
							if (pNew->pValue->pszKey)
								lstrcpy(pNew->pValue->pszKey, pszKey);
						}

						if (*ppNode != pParent)
							(*ppNode)->pNext = pNew;
						else
							pParent->pValue = pNew;

						pNext = pNew->pValue;
					}
					// Adding a new node to a node.
					else
					{
						if (!bKeyIsIndex)
						{
							if (*pszKey)
							{
								pNew->pszKey = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (lstrlen(pszKey) + 1));
								if (pNew->pszKey)
									lstrcpy(pNew->pszKey, pszKey);
							}
							else
							{
								pNew->pszKey = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR));
							}
						}

						if (*ppNode != pParent)
							(*ppNode)->pNext = pNew;
						else
							pParent->pValue = pNew;

						pNext = pNew;
					}
				}
			}

			GlobalFree(pszKey);
		}

		return pNext;
	}

	return NULL;
}

void JSON_Set(struct JSON_NODE* pNode, PBYTE pbValue, enum JSON_SET_FLAGS eFlags)
{
	if (pNode)
	{
		if (eFlags & JSF_IS_RAW)
		{
			JSON_Delete(pNode->pValue, NULL);

			pNode->eType = JNT_QUOTED_VALUE;

#ifdef UNICODE
			if (eFlags & JSF_IS_UNICODE)
			{
				pNode->pszValue = JSON_Quote((PWCHAR)pbValue, JQF_NO_QUOTE);
			}
			else
			{
				int cchValue = lstrlenA((PCHAR)pbValue);
				PWCHAR pwszConverted = JSON_ToUnicode((PCHAR)pbValue, &cchValue);
				if (pwszConverted)
				{
					pNode->pszValue = JSON_Quote(pwszConverted, JQF_NO_QUOTE);
					GlobalFree(pwszConverted);
				}
			}
#else
			if (eFlags & JSF_IS_UNICODE)
			{
				int cchValue = lstrlenW((PWCHAR)pbValue);
				PCHAR pszConverted = JSON_FromUnicode((PWCHAR)pbValue, &cchValue);
				if (pszConverted)
				{
					pNode->pszValue = JSON_Quote(pszConverted, JQF_NO_QUOTE);
					GlobalFree(pszConverted);
				}
			}
			else
			{
				pNode->pszValue = JSON_Quote((PCHAR)pbValue, JQF_NO_QUOTE);
			}
#endif
		}
		else
		{
			struct JSON_NODE* pChildNode = Parse(pbValue, eFlags);
			if (pChildNode)
			{
				switch (pNode->eType)
				{
				case JNT_NODE:

					// We are overwriting the node's value.
					JSON_Delete(pNode->pValue, NULL);

					switch (pChildNode->eType)
					{
					case JNT_NODE:
					case JNT_ARRAY:

						// Child node is a key: value pair; set it as the parent node's value.
						if (pChildNode->pszKey)
						{
							pNode->eType = JNT_NODE;
							pNode->pValue = pChildNode;
						}
						// Child node is a node without a key; overwrite the parent node's value with the child node's value.
						else
						{
							pNode->eType = pChildNode->eType;
							pNode->pValue = pChildNode->pValue;
							GlobalFree(pChildNode);
						}

						break;
					default:

						// Replace the node's value.
						pNode->eType = pChildNode->eType;
						pNode->pszValue = pChildNode->pszValue;
						GlobalFree(pChildNode);

						break;
					}

					break;
				case JNT_ARRAY:

					if (pNode->pValue)
					{
						struct JSON_NODE* pArrayNode = pNode->pValue;

						// Seek to the end of the array and then append to the end.
						while (TRUE)
						{
							if (pArrayNode->pNext == NULL)
								break;

							pArrayNode = pArrayNode->pNext;
						}

						pArrayNode->pNext = pChildNode;
					}
					else
					{
						pNode->pValue = pChildNode;
					}

					break;
				case JNT_VALUE:
				case JNT_QUOTED_VALUE:

					// We are overwriting the node's value. 
					if (pNode->pszValue)
						GlobalFree(pNode->pszValue);

					switch (pChildNode->eType)
					{
					case JNT_NODE:
					case JNT_ARRAY:

						// Child node is a key: value pair; set it as the parent node's value.
						if (pChildNode->pszKey)
						{
							pNode->eType = pChildNode->eType;
							pNode->pValue = pChildNode;
						}
						// Child node is a node without a key; overwrite the parent node's value with the child node's value.
						else
						{
							pNode->eType = pChildNode->eType;
							pNode->pValue = pChildNode->pValue;
							GlobalFree(pChildNode);
						}

						break;
					default:

						// Replace the node's value.
						pNode->eType = pChildNode->eType;
						pNode->pszValue = pChildNode->pszValue;
						GlobalFree(pChildNode);

						break;
					}

					break;
				}
			}
		}
	}
}

void JSON_SetEx(struct JSON_NODE* pNode, PTCHAR pszKey, BOOL bKeyIsIndex, PBYTE pbValue, enum JSON_SET_FLAGS eFlags)
{
	struct JSON_NODE* pNew = JSON_Next(&pNode, pszKey, bKeyIsIndex, TRUE);
	if (pNew)
		JSON_Set(pNew, pbValue, eFlags);
}

static void MyWriteFile(HANDLE hFile, PTCHAR pszText, int cchText, BOOL bAsUnicode)
{
	DWORD dwBytes;

#ifdef UNICODE
	if (bAsUnicode)
	{
		WriteFile(hFile, pszText, sizeof(WCHAR) * cchText, &dwBytes, NULL);
	}
	else
	{
		PCHAR pszConverted = JSON_FromUnicode(pszText, &cchText);
		if (pszConverted)
		{
			WriteFile(hFile, pszConverted, cchText, &dwBytes, NULL);
			GlobalFree(pszConverted);
		}
	}
#else
	if (bAsUnicode)
	{
		PWCHAR pwszConverted = JSON_ToUnicode(pszText, &cchText);
		if (pwszConverted)
		{
			WriteFile(hFile, pwszConverted, sizeof(WCHAR) * cchText, &dwBytes, NULL);
			GlobalFree(pwszConverted);
		}
	}
	else
	{
		WriteFile(hFile, pszText, cchText, &dwBytes, NULL);
	}
#endif
}

static void MyWriteFileEx(HANDLE hFile, PTCHAR pszText, int cchText, BOOL bAsUnicode, int iRepeat)
{
	int i = 0;

	for (; i < iRepeat; i++)
	{
		MyWriteFile(hFile, pszText, cchText, bAsUnicode);
	}
}

static void SerializeToFile(HANDLE hFile, struct JSON_NODE* pNode, int iIndent, BOOL bAsUnicode)
{
	if (!pNode)
		return;

	if (pNode->pszKey)
	{
		MyWriteFile(hFile, TEXT("\""), 1, bAsUnicode);
		MyWriteFile(hFile, pNode->pszKey, lstrlen(pNode->pszKey), bAsUnicode);
		MyWriteFile(hFile, TEXT("\":"), 2, bAsUnicode);

		if (iIndent)
		{
			MyWriteFile(hFile, TEXT(" "), 1, bAsUnicode);
		}
	}

	switch (pNode->eType)
	{
	case JNT_NODE:
	case JNT_ARRAY:
    
		MyWriteFile(hFile, pNode->eType == JNT_ARRAY ? TEXT("[") : TEXT("{"), 1, bAsUnicode);
    
		if (iIndent)
		{
			MyWriteFile(hFile, TEXT("\r\n"), 2, bAsUnicode);
			MyWriteFileEx(hFile, TEXT(JSON_INDENT_CHAR), 1, bAsUnicode, iIndent);
		}

		SerializeToFile(hFile, pNode->pValue, iIndent ? iIndent + JSON_INDENT : 0, bAsUnicode);
    
		if (iIndent)
		{
			MyWriteFile(hFile, TEXT("\r\n"), 2, bAsUnicode);
			MyWriteFileEx(hFile, TEXT(JSON_INDENT_CHAR), 1, bAsUnicode, iIndent - 1);
		}

		MyWriteFile(hFile, pNode->eType == JNT_ARRAY ? TEXT("]") : TEXT("}"), 1, bAsUnicode);

		break;
	case JNT_VALUE:

		MyWriteFile(hFile, pNode->pszValue, lstrlen(pNode->pszValue), bAsUnicode);

		break;
	case JNT_QUOTED_VALUE:
    
		MyWriteFile(hFile, TEXT("\""), 1, bAsUnicode);
		MyWriteFile(hFile, pNode->pszValue, lstrlen(pNode->pszValue), bAsUnicode);
		MyWriteFile(hFile, TEXT("\""), 1, bAsUnicode);
    
		break;
	}

	if (pNode->pNext)
	{
		pNode = pNode->pNext;
    
		MyWriteFile(hFile, TEXT(","), 1, bAsUnicode);
    
		if (iIndent)
		{
			MyWriteFile(hFile, TEXT("\r\n"), 2, bAsUnicode);
			MyWriteFileEx(hFile, TEXT(JSON_INDENT_CHAR), 1, bAsUnicode, iIndent - 1);
		}

		SerializeToFile(hFile, pNode, iIndent, bAsUnicode);
	}
}

static void MyStrCpy(PTCHAR* ppszBuffer, int* pcchBuffer, int* piPos, PTCHAR pszCopy, int cchCopyLen, BOOL bResize)
{
	if (*piPos + cchCopyLen <= *pcchBuffer)
	{
		if (ppszBuffer)
			lstrcpy(*ppszBuffer + *piPos, pszCopy);
		*piPos += cchCopyLen;
	}
	else if (bResize)
	{
		*pcchBuffer *= 2;
		if (ppszBuffer)
			*ppszBuffer = (PTCHAR)GlobalReAlloc(*ppszBuffer, *pcchBuffer, GMEM_ZEROINIT);
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pszCopy, cchCopyLen, TRUE);
	}
	else
	{
		int i = 0;
		for (; i < cchCopyLen && *piPos < *pcchBuffer; i++, (*piPos)++)
		{
			if (ppszBuffer)
				(*ppszBuffer)[*piPos] = pszCopy[i];
		}
	}
}

static void MyStrCpyEx(PTCHAR* ppszBuffer, int* pcchBuffer, int* piPos, PTCHAR pszCopy, int cchCopyLen, int iRepeat, BOOL bResize)
{
	int i = 0;
	for (; i < iRepeat; i++)
	{
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pszCopy, cchCopyLen, bResize);
	}
}

static void SerializeToBuffer(PTCHAR* ppszBuffer, int* pcchBuffer, struct JSON_NODE* pNode, int* piPos, int iIndent, BOOL bResize)
{
	if (!pNode)
	{
		if (ppszBuffer)
			lstrcpy(*ppszBuffer + *piPos, TEXT(""));
		return;
	}

	if (pNode->pszKey)
	{
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT("\""), 1, bResize);
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pNode->pszKey, lstrlen(pNode->pszKey), bResize);
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT("\":"), 2, bResize);

		if (iIndent)
		{
			MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT(" "), 1, bResize);
		}
	}

	switch (pNode->eType)
	{
	case JNT_NODE:
	case JNT_ARRAY:
    
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pNode->eType == JNT_ARRAY ? TEXT("[") : TEXT("{"), 1, bResize);

		if (iIndent)
		{
			MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT("\r\n"), 2, bResize);
			MyStrCpyEx(ppszBuffer, pcchBuffer, piPos, TEXT(JSON_INDENT_CHAR), 1, iIndent, bResize);
		}

		SerializeToBuffer(ppszBuffer, pcchBuffer, pNode->pValue, piPos, iIndent ? iIndent + JSON_INDENT : 0, bResize);

		if (iIndent)
		{
			MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT("\r\n"), 2, bResize);
			MyStrCpyEx(ppszBuffer, pcchBuffer, piPos, TEXT(JSON_INDENT_CHAR), 1, iIndent - 1, bResize);
		}

		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pNode->eType == JNT_ARRAY ? TEXT("]") : TEXT("}"), 1, bResize);

		break;
	case JNT_VALUE:
    
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pNode->pszValue, lstrlen(pNode->pszValue), bResize);

		break;
	case JNT_QUOTED_VALUE:
    
		MyStrCpy(ppszBuffer, pcchBuffer, piPos,  TEXT("\""), 1, bResize);
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, pNode->pszValue, lstrlen(pNode->pszValue), bResize);
		MyStrCpy(ppszBuffer, pcchBuffer, piPos,  TEXT("\""), 1, bResize);
    
		break;
	}

	if (pNode->pNext)
	{
		pNode = pNode->pNext;
    
		MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT(","), 1, bResize);

		if (iIndent)
		{
			MyStrCpy(ppszBuffer, pcchBuffer, piPos, TEXT("\r\n"), 2, bResize);
			MyStrCpyEx(ppszBuffer, pcchBuffer, piPos, TEXT(JSON_INDENT_CHAR), 1, iIndent - 1, bResize);
		}

		SerializeToBuffer(ppszBuffer, pcchBuffer, pNode, piPos, iIndent, bResize);
	}
}

static BOOL AddRoot(struct JSON_NODE** ppNode)
{
	struct JSON_NODE* pRoot = JSON_Create();
	if (pRoot)
	{
		pRoot->eType = JNT_NODE;
		pRoot->pValue = *ppNode;
		*ppNode = pRoot;
		return TRUE;
	}
	
	return FALSE;
}

static TCHAR i2a(TCHAR code)
{
	return hex[code & 15];
}
 
static int EscapePostData(PTCHAR* ppszPostData, int* pcchPostData)
{
	int cchLength = 0;
	PTCHAR pszBuffer = (PTCHAR)GlobalAlloc(GPTR, (lstrlen(*ppszPostData) * 3 + 1) * sizeof(TCHAR));

	if (pszBuffer)
	{
		PTCHAR pszBufferPtr = pszBuffer, pszPostDataPtr = *ppszPostData;
		int iPos = 0;

		while (*pszPostDataPtr)
		{
			if (IsCharAlphaNumeric(*pszPostDataPtr) || *pszPostDataPtr == TEXT('-') || *pszPostDataPtr == TEXT('_') || *pszPostDataPtr == TEXT('.') || *pszPostDataPtr == TEXT('~'))
			{
				*pszBufferPtr++ = *pszPostDataPtr;
			}
			else if (*pszPostDataPtr == TEXT(' '))
			{
				*pszBufferPtr++ = TEXT('+');
			}
			else
			{
				*pszBufferPtr++ = TEXT('%');
				*pszBufferPtr++ = i2a(*pszPostDataPtr >> 4);
				*pszBufferPtr++ = i2a(*pszPostDataPtr & 15);
			}

			pszPostDataPtr++;
		}

		if ((cchLength = lstrlen(pszBuffer)) > 0)
			MyStrCpy(ppszPostData, pcchPostData, &iPos, pszBuffer, cchLength, TRUE);
	}

	return cchLength;
}

BOOL JSON_Serialize(struct JSON_NODE* pNode, PTCHAR pszBuffer, int cchBuffer, BOOL bIsFile, BOOL bAsUnicode, BOOL bFormat)
{
	BOOL bResult = FALSE;

	BOOL bAddRoot = pNode && pNode->pszKey && *pNode->pszKey;
	if (bAddRoot && !AddRoot(&pNode))
		bAddRoot = FALSE;

	if (bIsFile)
	{
		HANDLE hFile = CreateFile(pszBuffer, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile != INVALID_HANDLE_VALUE)
		{
			SerializeToFile(hFile, pNode, bFormat ? JSON_INDENT : 0, bAsUnicode);
			CloseHandle(hFile);
			bResult = TRUE;
		}
	}
	else
	{
		int iPos = 0;
		SerializeToBuffer(&pszBuffer, &cchBuffer, pNode, &iPos, bFormat ? JSON_INDENT : 0, FALSE);
		bResult = TRUE;
	}

	if (bAddRoot)
		GlobalFree(pNode);

	return bResult;
}

PTCHAR JSON_SerializeAlloc(struct JSON_NODE* pNode, BOOL bFormat, BOOL bAsPostData)
{
	int cchBuffer = 2048;
	PTCHAR pszBuffer = (PTCHAR)GlobalAlloc(GPTR, cchBuffer * sizeof(TCHAR));
	
	if (pszBuffer)
	{
		BOOL bAddRoot = pNode && pNode->pszKey && *pNode->pszKey;
		if (bAddRoot && !AddRoot(&pNode))
			bAddRoot = FALSE;

		if (bAsPostData)
		{
			struct JSON_NODE* pNext = pNode->pValue;
			int iPos = 0;
			
			while (pNext)
			{
				switch (pNext->eType)
				{
				case JNT_ARRAY:
					{
						struct JSON_NODE* pArrayNext = pNext->pValue;
						while (pArrayNext)
						{
							MyStrCpy(&pszBuffer, &cchBuffer, &iPos, pNext->pszKey, lstrlen(pNext->pszKey), TRUE);
							MyStrCpy(&pszBuffer, &cchBuffer, &iPos, TEXT("[]="), 3, TRUE);

							switch (pArrayNext->eType)
							{
							case JNT_NODE:
							case JNT_ARRAY:
								SerializeToBuffer(&pszBuffer, &cchBuffer, pArrayNext, &iPos, 0, TRUE);
								break;
							default:
								MyStrCpy(&pszBuffer, &cchBuffer, &iPos, pArrayNext->pszValue, lstrlen(pArrayNext->pszValue), TRUE);
								break;
							}

							pArrayNext = pArrayNext->pNext;
							if (pArrayNext)
							  MyStrCpy(&pszBuffer, &cchBuffer, &iPos, TEXT("&"), 1, TRUE);
						}
					}
					break;
				default:
					{
						int cchValueBuffer = 2048;
						PTCHAR pszValueBuffer = (PTCHAR)GlobalAlloc(GPTR, cchValueBuffer * sizeof(TCHAR));

						MyStrCpy(&pszBuffer, &cchBuffer, &iPos, pNext->pszKey, lstrlen(pNext->pszKey), TRUE);
						if (pNext->eType != JNT_VALUE || lstrcmpi(pNext->pszValue, TEXT("true")) != 0)
						{
							MyStrCpy(&pszBuffer, &cchBuffer, &iPos, TEXT("="), 1, TRUE);

							if (pszValueBuffer)
							{
								int iValuePos = 0;

								if (pNext->eType == JNT_NODE)
									SerializeToBuffer(&pszValueBuffer, &cchValueBuffer, pNext->pValue, &iValuePos, bFormat ? JSON_INDENT : 0, TRUE);
								else
									MyStrCpy(&pszValueBuffer, &cchValueBuffer, &iValuePos, pNext->pszValue, lstrlen(pNext->pszValue), TRUE);

								if (iValuePos = EscapePostData(&pszValueBuffer, &cchValueBuffer))
									MyStrCpy(&pszBuffer, &cchBuffer, &iPos, pszValueBuffer, iValuePos, TRUE);

								GlobalFree(pszValueBuffer);
							}
						}
					}
					break;
				}

				pNext = pNext->pNext;
				if (pNext)
					MyStrCpy(&pszBuffer, &cchBuffer, &iPos, TEXT("&"), 1, TRUE);
			}
		}
		else
		{
			int iPos = 0;
			SerializeToBuffer(&pszBuffer, &cchBuffer, pNode, &iPos, bFormat ? JSON_INDENT : 0, TRUE);
		}

		if (bAddRoot)
			GlobalFree(pNode);
	}

	return pszBuffer;
}

PTCHAR JSON_Expand(struct JSON_NODE* pNode)
{
	PTCHAR pszExpanded = NULL;
	int i, j, cch = lstrlen(pNode->pszValue);
	TCHAR szUnicode[7];

	pszExpanded = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (cch + 1));
	if (pszExpanded)
	{
		for (i = 0, j = 0; i < cch; i++, j++)
		{
			if (pNode->pszValue[i] == TEXT('\\'))
			{
				switch (pNode->pszValue[i + 1])
				{
				case TEXT('"'):
				case TEXT('\\'):
				case TEXT('/'):
					pszExpanded[j] = pNode->pszValue[i + 1];
					i++;
					break;
				case TEXT('b'):
					pszExpanded[j] = TEXT('\b');
					i++;
					break;
				case TEXT('f'):
					pszExpanded[j] = TEXT('\f');
					i++;
					break;
				case TEXT('n'):
					pszExpanded[j] = TEXT('\n');
					i++;
					break;
				case TEXT('r'):
					pszExpanded[j] = TEXT('\r');
					i++;
					break;
				case TEXT('t'):
					pszExpanded[j] = TEXT('\t');
					i++;
					break;
				case TEXT('u'):
					wsprintf(szUnicode, TEXT("0x%c%c%c%c"), pNode->pszValue[i + 2], pNode->pszValue[i + 3], pNode->pszValue[i + 4], pNode->pszValue[i + 5]);
#ifdef UNICODE
					pszExpanded[j] = (WCHAR)myatoi(szUnicode);
#else
					wsprintfW((PWCHAR)szUnicode, L"%c", myatoi(szUnicode));
					if (WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, (PWCHAR)szUnicode, 1, NULL, 0, NULL, NULL) == 1)
						WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, (PWCHAR)szUnicode, 1, pszExpanded + j, 1, NULL, NULL);
					else
						lstrcpyA(pszExpanded + j, "?");
#endif
					i += 5;
					break;
				}
			}
			else
			{
				pszExpanded[j] = pNode->pszValue[i];
			}
		}
	}

	return pszExpanded;
}

PTCHAR JSON_Quote(PTCHAR pszValue, enum JSON_QUOTE_FLAGS eFlags)
{
	PTCHAR pszQuoted;
	int i, cch = lstrlen(pszValue), cchNew = cch;
	BOOL bQuote = !(eFlags & JQF_NO_QUOTE) && ((eFlags & JQF_ALWAYS_QUOTE) || cch < 2 || cch > 1 && (pszValue[0] != TEXT('"') || pszValue[cch - 1] != TEXT('"')));
	if (bQuote)
		cchNew += 2;
	
	for (i = 0; i < cch; i++)
	{
		switch (pszValue[i])
		{
		case TEXT('\b'):
		case TEXT('\f'):
		case TEXT('\n'):
		case TEXT('\r'):
		case TEXT('\t'):
		case TEXT('\\'):
		case TEXT('"'):
			cchNew++;
			break;
		default:
			if ((eFlags & JQF_ESCAPE_UNICODE) && (pszValue[i] < 32 || pszValue[i] > 126))
				cchNew += 6;
			break;
		}
	}

	pszQuoted = (PTCHAR)GlobalAlloc(GPTR, sizeof(TCHAR) * (cchNew + 1));

	if (pszQuoted && cch < cchNew)
	{
		int j;
		
		if (bQuote)
		{
			i = 0;
		}
		else
		{
			i = 1;
			cch--;
		}
		
		pszQuoted[0] = TEXT('"');

		for (j = 1; i < cch; i++, j++)
		{
			switch (pszValue[i])
			{
			case TEXT('\b'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = TEXT('b');
				break;
			case TEXT('\f'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = TEXT('f');
				break;
			case TEXT('\n'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = TEXT('n');
				break;
			case TEXT('\r'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = TEXT('r');
				break;
			case TEXT('\t'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = TEXT('t');
				break;
			case TEXT('\\'):
			case TEXT('"'):
				pszQuoted[j++] = TEXT('\\');
				pszQuoted[j] = pszValue[i];
				break;
			default:
				if ((eFlags & JQF_ESCAPE_UNICODE) && (pszValue[i] < 32 || pszValue[i] > 126))
				{
					j += wsprintf(pszQuoted + j, TEXT("\\u%04x"), (unsigned char)pszValue[i]) - 1;
				}
				else
				{
					pszQuoted[j] = pszValue[i];
				}
				break;
			}
		}
		
		pszQuoted[j] = TEXT('"');
	}
	else
	{
		lstrcpy(pszQuoted, pszValue);
	}

	return pszQuoted;
}

PCHAR JSON_FromUnicode(PWCHAR pwszText, int* pcchText)
{
	int cbConverted = WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, pwszText, *pcchText, NULL, 0, NULL, NULL);
	if (cbConverted > 0)
	{
		PCHAR pszBuffer = (PCHAR)GlobalAlloc(GPTR, sizeof(CHAR) * (cbConverted + 1));
		if (pszBuffer)
		{
			if (WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, pwszText, *pcchText, pszBuffer, cbConverted, NULL, NULL) > 0)
			{
				*pcchText = cbConverted;
				return pszBuffer;
			}

			GlobalFree(pszBuffer);
		}
	}

	return NULL;
}

PWCHAR JSON_ToUnicode(PCHAR pszText, int* pcbText)
{
	int cchConverted = MultiByteToWideChar(CP_ACP, MB_COMPOSITE, pszText, *pcbText, NULL, 0);
	if (cchConverted > 0)
	{
		PWCHAR pwszBuffer = (PWCHAR)GlobalAlloc(GPTR, sizeof(WCHAR) * (cchConverted + 1));
		if (pwszBuffer)
		{
			if (MultiByteToWideChar(CP_ACP, MB_COMPOSITE, pszText, *pcbText, pwszBuffer, cchConverted) > 0)
			{
				*pcbText = cchConverted;
				return pwszBuffer;
			}

			GlobalFree(pwszBuffer);
		}
	}

	return NULL;
}